(* $Id: godi_tty.mli 804 2008-06-08 22:21:03Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* Access to character terminals
 *
 * This module allows it to access the properties of character terminals,
 * and to poll for keyboard and other important events. There may be
 * several implementations for this module, depending on the target
 * platform.
 *)

(* Output *)

val out_init : unit -> unit
  (* Reinitializes the output *)

val out_string : string -> unit
  (* Prints the string. NO control codes are interpreted *)

val out_home : unit -> unit
  (* Set cursor to upper left corner *)

val out_clear : unit -> unit
  (* Clear screen, cursor at top left corner *)

val out_clear_eol : unit -> unit
  (* Clear until end of line *)

val out_to_bol : unit -> unit
  (* Go to beginning of current line *)

val out_to_bol_next : unit -> unit
  (* Go to beginnning of next line *)

val lines : unit -> int
  (* Number of lines *)

val columns : unit -> int 
  (* Number of columns *)


(* Input *)

type key =
  [ `Return
  | `BS
  | `ESC
  | `PageUp
  | `PageDown
  | `Beginning
  | `End
  | `Char of char
  | `None
  ]

val poll_keyboard : ?wait:bool -> unit -> key
  (* Checks whether the user has pressed a key, and returns the key.
   * If no key has been pressed, or if the system is currently unable to
   * determine this, the value `None will be returned. By default, the
   * function returns immediately with `None when there is no keypress.
   * There may be other reasons to return `None. For example, on Unix
   * systems the SIGCONT signal also causes that `None is returned.
   *
   * @param wait If false (the default), it is not waited for a keypress.
   * If true, however, the routine waits until there is a keypress to
   * report. It is allowed to return `None in special cases.
   *)

val load_script : Godi_file.filename -> unit
  (* Loads a script simulating key presses. The contents of the
   * script are returned by the following [poll_keyboard] invocations.
   * The file is read line by line. The lines may either contain
   * exactly one keyword, or a series of characters. The linefeed
   * characters at the end of the lines are ignored.
   * Keywords are: RETURN, BS, ESC, PGUP, PGDOWN, BEGINNING, END
   *)

val forced_refresh_event : unit -> float
  (* Returns the Sys.time of the last system event for which the program
   * must completely repaint the screen. On Unix systems, the SIGCONT
   * signal is such an event.
   *)
